
# THIS FILE IS GENERATED FROM ORANGE SETUP.PY
short_version = '3.3.12'
version = '3.3.12'
full_version = '3.3.12'
git_revision = '2385b94eb5491909bd216597ad5bde79be1beb86'
release = True

if not release:
    version = full_version
    short_version += ".dev"
